from __future__ import print_function
from collections import namedtuple
from platform import platform
import os
import shutil
import subprocess
import sys
import time

try:
    import _winreg as winreg
except ImportError:
    import winreg

def _which_python():
    for dir in os.environ.get("Path").split(os.pathsep):
        full_path = os.path.join(dir, "python.exe")
        if os.path.exists(full_path):
            return full_path
    return None

BRCommand = namedtuple("BRCommand", ["bin", "args", "output"])

# Since buildout's python.exe is really a script passed to the
# system-installed Python, we need to get that same environment by starting
# Python in the same manner.
# Ex. ['c:\\Python27\\python.exe', 'C:\\u1\\bin\\python-script.py']
PYTHON = [_which_python()]

def _find_bitrock():
    with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,
            r"Software\BitRock\BitRock InstallBuilder Professional", 0,
            winreg.KEY_READ | winreg.KEY_WOW64_32KEY) as key:
        path, reg_type = winreg.QueryValueEx(key, "Location")
        if reg_type != winreg.REG_SZ:
            raise Exception(
                "Got unexpected Location information: {}".format(reg_type))
    return path if os.path.exists(path) else None

def _find_output_dir(type):
    if os.getenv("JENKINS"):
        output = os.path.join(_find_bitrock(), "output")
    else:
        user_dir = os.getenv("USERPROFILE")
        if not user_dir:
            print("No USERPROFILE available")
            return None
        dir = "My Documents" if "Windows-2008" in platform() else "Documents"
        output = os.path.join(user_dir, dir, type, "output")
    return output if os.path.exists(output) else None

def _get_last_file(dir):
    """Get the last written file in this directory. We just wrote it."""
    files = os.listdir(dir)
    # Sort the files by modification time.
    files.sort(key=lambda f: os.stat(os.path.join(dir, f)).st_mtime)
    # We want the most recently modified, so take the last one.
    return os.path.join(dir, files[-1])

def setup_command(cmd):
    """Run a command through the setup.py script"""
    print("Calling", cmd)
    rv = subprocess.Popen(PYTHON + ["setup.py", cmd]).wait()
    if rv != 0:
        raise Exception("{} returned {}".format(cmd, rv))
    print("{} returned {}".format(cmd, rv))

def bitrock_command(cmd, rename=False):
    """Run one of the BitRock binaries with some arguments"""

    if not cmd.output:
        raise Exception("Unable to find {} output".format(cmd.bin))

    print("Running BitRock", cmd.bin)
    subprocess.check_output(
            [os.path.join(_find_bitrock(), cmd.bin)] + cmd.args,
            stderr=subprocess.STDOUT)

    source = _get_last_file(cmd.output)
    target = os.path.join(os.path.dirname(__file__) or os.getcwd(),
                          os.path.basename(source))

    # Copy the installer locally
    shutil.copy(source, target)
    if rename:
        orig_target = target
        # Rename the installer to be timestamped.
        root, ext = os.path.splitext(target)
        timestamp = time.strftime("%Y%m%d-%H%M%S", time.gmtime())
        target = "{}-{}{}".format(root, timestamp, ext)
        os.rename(orig_target, target)
    print("Created", target)

def main():
    for cmd in ["fetch", "prepare", "py2exe"]:
        setup_command(cmd)

    installer = BRCommand(os.path.join("bin", "builder-cli.exe"),
                          ["build", "ubuntuone.xml"],
                          _find_output_dir("InstallBuilder"))
    bitrock_command(installer, True)

if __name__ == "__main__":
    main()

