# setup-mac.py - Build system for Ubuntu One Client package
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Setup.py script for Ubuntu One Mac Client

Usage:
    % python setup.py prepare    # builds and arranges subpackages

    use --from-trunk to grab sources from launchpad. otherwise, builds
    from buildout parts dir.

    use --only_prepare='sso, controlpanel, client' (or 'sso, client',
    etc) to limit which of those three to stage. This is only useful
    for development

    % python setup.py py2app # builds .app
"""


import copy
import errno
import glob
import os
import shutil
import subprocess
import sys
from zipfile import ZipFile

import conf

from distutils.cmd import Command
from distutils.core import setup
from plistlib import Plist

try:
    import py2app
except ImportError:
    print "This setup requires py2app, check your path"
    print "or see http://bitbucket.com/ronaldoussoren/py2app/"
    sys.exit()

assert(py2app)

# NOTE - this needs to be the full CN, not just a substring.
# /usr/bin/codesign will find the right cert if you pass it a
# substring, but the rest of the system wants an exact match.
CODESIGN_CN = "Mac Developer: Michael McCracken (GP72FH8MSU)"

FSEVENTS_DAEMON_NAME = "com.ubuntu.one.fsevents"

APP_NAMES = {'ubuntu-sso-login': 'Ubuntu SSO Helper',
             'ubuntu-sso-login-qt': 'Ubuntu Single Sign-On',
             'ubuntu-sso-ssl-certificate-qt': 'Ubuntu SSO SSL Certificate',
             'ubuntu-sso-proxy-creds-qt': 'Ubuntu SSO Proxy Credentials',
             'ubuntuone-syncdaemon': 'UbuntuOne Syncdaemon',
             'ubuntuone-control-panel-qt': 'Ubuntu One',
             'ubuntuone-proxy-tunnel': 'UbuntuOne Proxy Tunnel',
             'ubuntuone-updater': 'UbuntuOne Updater',
             'u1sdtool': 'U1SDTool'
             }

LOG_LEVEL = "DEBUG"
LOG_FILE_SIZE = "1000000000"

INSTALL_DIR = os.path.abspath("installed")

# bzr env assumes homebrew install:
# This is necessary so we can run setup-mac with the buildout python,
# which breaks the paths needed for bzr.
# just grab the latest installed bzr:
BZR_PATHS = sorted(glob.glob("/usr/local/Cellar/bazaar/*/libexec/"),
                   reverse=True)
if len(BZR_PATHS) < 1:
    print "Could not find bzr in /usr/local/Cellar/bazaar."
    sys.exit()

BZR_PATH = BZR_PATHS[0] + ":/usr/local/bin:/usr/bin"
BZR_ENV = os.environ.copy()
BZR_ENV.update(dict(PYTHONPATH=BZR_PATHS[0],
                    PATH=BZR_PATH))


def branch_and_merge(urls):
    """Branch and merge all requested branches."""
    project_name = urls[0].split(":")[1]
    print "Getting sources for:", project_name
    folder_name = os.path.join('sources', project_name)
    current_folder = os.getcwd()
    if os.path.isdir(folder_name):  # Folder exists
        shutil.rmtree(folder_name)
    subprocess.check_call(["bzr", "branch", urls[0], folder_name],
                          env=BZR_ENV)
    os.chdir(folder_name)
    for url in urls[1:]:
        print "Merging:", url
        subprocess.check_call(["bzr", "merge", url], env=BZR_ENV)
        subprocess.check_call(["bzr", "commit", "-m", '"merged %s"' % url],
                              env=BZR_ENV)
    os.chdir(current_folder)


# pylint is complaining about parent classes having too many methods
# pylint: disable=R0904
class Clean(Command):
    """Remove built products from previous runs."""

    user_options = []

    def initialize_options(self):
        """No-op."""

    def finalize_options(self):
        """No-op."""

    def run(self):
        """Clean up."""
        for d in [INSTALL_DIR, 'dist', 'build']:
            if os.path.exists(d):
                print "Clean: removing", d
                shutil.rmtree(d)
        print "NOTE: exiting. 'clean' cannot be chained, "
        print "      because we call setup() multiple times."
        sys.exit()

# pylint: enable=R0904


# pylint is complaining about parent classes having too many methods
# pylint: disable=R0904
class DoNotPrepareAnything(Command):
    """No-op, lets us call setup multiple times without re-preparing."""

    user_options = [("from-trunk", None, "noop."),
                    ("only-prepare=", None, "noop."),
                    ("source-dir=", None, "noop.")]

    def initialize_options(self):
        """No-op."""
        self.only_prepare = None
        self.from_trunk = None
        self.source_dir = None

    def finalize_options(self):
        """No-op."""

    def run(self):
        """No-op."""

# pylint: enable=R0904


# pylint is complaining about parent classes having too many methods
# pylint: disable=R0904
class PrepareSources(Command):
    """Manipulate the sources to look as if they were installed."""

    user_options = [("from-trunk", None,
                     """Download sources from launchpad. default is to
                     use buildout instead."""),
                    ("only-prepare=", None,
                     """Only do expensive stuff for the packages on
                     this list. Can be any of: 'sso',
                     'controlpanel', 'client', or 'storage-protocol'.
                     (NOTE: still downloads everything if from-trunk
                     is set)"""),
                    ("source-dir=", None,
                     """Set the directory to get the sources from.
                     The default depends on from-trunk:
                     * if from-trunk is not specified,
                       the default is buildout sources dir.
                     * if from-trunk is specified,
                       the default source dir is './sources'""")]

    def initialize_options(self):
        """Init options."""
        self.from_trunk = None
        self.only_prepare = None
        self.source_dir = None

    def finalize_options(self):
        """Finalize options."""
        if self.source_dir is None:
            if self.from_trunk is not None:
                self.source_dir = 'sources'
            else:
                self.source_dir = os.path.join(os.path.dirname(__file__),
                                               "devsetup", "parts")

        if self.only_prepare is None:
            # default is to do everything:
            self.only_prepare = "sso, controlpanel, client, storage-protocol"

    def copy_translation_files(self, src_root, dst_root):
        """Merge .mo files from src_root into dst_root."""
        for src_dir, dirs, files in os.walk(src_root):
            dst_dir = src_dir.replace(src_root, dst_root)
            if not os.path.exists(dst_dir):
                os.mkdir(dst_dir)
            for f in files:
                src_f = os.path.join(src_dir, f)
                dst_f = os.path.join(dst_dir, f)
                if os.path.exists(dst_f):
                    raise Exception("Error copying translations: %r "
                                    "exists already" % dst_f)
                shutil.copyfile(src_f, dst_f)

    def run(self):
        """Copy, build and munge things to prepare for packaging."""

        print "Starting Prepare step:"

        for folder in ["bin", INSTALL_DIR, "data", "build"]:
            if not os.path.isdir(folder):
                os.mkdir(folder)

        if self.from_trunk is not None:
            if not os.path.isdir(self.source_dir):
                os.mkdir(self.source_dir)

            branch_and_merge(conf.U1_CLIENT_BRANCHES)
            branch_and_merge(conf.U1_CLIENT_DATA_BRANCHES)
            branch_and_merge(conf.U1_CONTROL_PANEL_BRANCHES)
            branch_and_merge(conf.UBUNTU_SSO_BRANCHES)
            branch_and_merge(conf.U1_STORAGE_PROTOCOL_BRANCHES)

        print "will look for sources in ", self.source_dir
        print "and install into ", INSTALL_DIR

        # Copy main executables.  In some cases, add .py so py2app
        # will recognize the file as python and not crash.

        shutil.copyfile(os.path.join(self.source_dir,
                                     "ubuntuone-client", "bin",
                                     "ubuntuone-syncdaemon"),
                        os.path.join("bin", "ubuntuone-syncdaemon.py"))

        shutil.copyfile(os.path.join(self.source_dir, "ubuntuone-client",
                                     "bin", "u1sdtool"),
                        os.path.join("bin", "u1sdtool.py"))

        dest = os.path.join("bin", "ubuntuone-proxy-tunnel.py")
        shutil.copyfile(os.path.join(self.source_dir, "ubuntuone-client",
                                     "bin", "ubuntuone-proxy-tunnel"),
                        dest)

        dest = os.path.join("bin", "ubuntu-sso-login.py")
        shutil.copyfile(os.path.join(self.source_dir, "ubuntu-sso-client",
                                     "bin", "ubuntu-sso-login"), dest)
        os.chmod(dest, 0755)

        dest = os.path.join("bin", "ubuntu-sso-login-qt.py")
        shutil.copyfile(os.path.join(self.source_dir, "ubuntu-sso-client",
                                     "bin", "ubuntu-sso-login-qt"), dest)

        shutil.copyfile(os.path.join(self.source_dir, "ubuntu-sso-client",
                                     "bin", "ubuntu-sso-proxy-creds-qt"),
                        os.path.join("bin", "ubuntu-sso-proxy-creds-qt.py"))

        shutil.copyfile(os.path.join(self.source_dir, "ubuntu-sso-client",
                                     "bin", "ubuntu-sso-ssl-certificate-qt"),
                        os.path.join("bin",
                                     "ubuntu-sso-ssl-certificate-qt.py"))

        shutil.copyfile(os.path.join(self.source_dir,
                                     "ubuntuone-control-panel", "bin",
                                     "ubuntuone-updater"),
                        os.path.join("bin", "ubuntuone-updater.py"))

        shutil.copyfile(os.path.join(self.source_dir,
                                     "ubuntuone-control-panel", "bin",
                                     "ubuntuone-control-panel-qt"),
                        os.path.join("bin", "ubuntuone-control-panel-qt.py"))

        # Remove "installed" copy
        try:
            shutil.rmtree(os.path.join(INSTALL_DIR, "lib"))
        except OSError, e:
            error, msg = e
            if error != errno.ENOENT:
                print ("WARNING: got OSError %r in rmtree(%r)."
                       % (e, os.path.join(INSTALL_DIR, "lib")))

        start_dir = os.getcwd()

        # Build SSO UI files, copy packages.
        dest_dir = os.path.join(INSTALL_DIR, "lib", "site-packages")
        dest_sso = os.path.join(dest_dir, "ubuntu_sso")

        os.chdir(os.path.join(self.source_dir, "ubuntu-sso-client"))

        if "sso" in self.only_prepare:
            print "building and installing ubuntu-sso-client"
            print "from directory ", os.getcwd()
            log_file_name = os.path.join(INSTALL_DIR, "sso-build.log")
            with open(log_file_name, 'w') as logfile:
                cmd = "python setup.py build"
                retval = subprocess.call(cmd,
                                         shell=True,
                                         stderr=subprocess.STDOUT,
                                         stdout=logfile
                                         )
                if retval != 0:
                    print "error running %r in %r. look in %r for details" %\
                          (cmd, os.getcwd(), logfile.name)

            # install UI files from sso. ignore the ubuntu_sso module
            # this installs:
            log_file_name = os.path.join(INSTALL_DIR, "sso-install.log")
            with open(log_file_name, 'w') as logfile:
                cmd = "python setup.py install --prefix=%s" % INSTALL_DIR,
                retval = subprocess.call(cmd,
                                         shell=True,
                                         stderr=subprocess.STDOUT,
                                         stdout=logfile
                                         )
                if retval != 0:
                    print "error running %r in %r. look in %r for details" %\
                          (cmd, os.getcwd(), logfile.name)

            # copy ubuntu_sso module separately to
            # install/lib/site-packages/ubuntu_sso
            try:
                shutil.copytree("ubuntu_sso", dest_sso)
            except Exception, e:
                print e
                print "tried to copy", os.path.abspath('ubuntu_sso'),
                print "to ", dest_sso


        sso_revno = subprocess.check_output(["bzr", "revno", "."],
                                            env=BZR_ENV).strip()

        os.chdir(start_dir)

        # install u1 client package to install/lib/site-packages/ubuntuone:
        os.chdir(os.path.join(self.source_dir, "ubuntuone-client"))

        if "client" in self.only_prepare:
            print "copying ubuntuone-client package"
            shutil.copy(os.path.join("windows", "clientdefs.py"),
                os.path.join("ubuntuone", "clientdefs.py"))
            dest_client = os.path.join(dest_dir, "ubuntuone")
            shutil.copytree("ubuntuone", dest_client)

            # workaround for py2app issue #49:
            # (it can't handle non-py files in a py module directory)
            os.remove(os.path.join(dest_client,
                                   "syncdaemon",
                                   "u1fsfsm.ods"))

        u1client_revno = subprocess.check_output(["bzr", "revno", "."],
                                                 env=BZR_ENV).strip()

        os.chdir(start_dir)

        # install controlpanel files
        cp_source_dir = os.path.join(self.source_dir,
                                     "ubuntuone-control-panel")
        os.chdir(cp_source_dir)

        if "controlpanel" in self.only_prepare:
            print "building ubuntuone-controlpanel"
            log_file_name = os.path.join(INSTALL_DIR, "u1cp-build.log")
            with open(log_file_name, 'w') as logfile:
                cmd = "python setup.py build"
                retval = subprocess.call(cmd,
                                         shell=True,
                                         stderr=subprocess.STDOUT,
                                         stdout=logfile
                                         )
                if retval != 0:
                    print "error running %r in %r. look in %r for details" %\
                          (cmd, os.getcwd(), logfile.name)

            # Copying by hand because the install is borked
            dest_cp = os.path.join(dest_dir, "ubuntuone", "controlpanel")
            shutil.copytree(os.path.join("ubuntuone", "controlpanel"),
                            dest_cp)

            # copy translations over: (we are already in cp_source_dir)
            self.copy_translation_files(os.path.join('build', 'mo'),
                                        os.path.join(INSTALL_DIR,
                                                     'share', 'locale'))

        u1cp_revno = subprocess.check_output(["bzr", "revno", "."],
                                             env=BZR_ENV).strip()

        os.chdir(start_dir)

        # build and install storage-protocol files
        os.chdir(os.path.join(self.source_dir, "ubuntuone-storage-protocol"))
        if 'storage-protocol' in self.only_prepare:
            print "building ubuntuone-storage-protocol"
            log_file_name = os.path.join(INSTALL_DIR,
                                         "storage-protocol-build.log")
            with open(log_file_name, 'w') as logfile:
                cmd = "python setup.py build"
                retval = subprocess.call(cmd,
                                         shell=True,
                                         stderr=subprocess.STDOUT,
                                         stdout=logfile
                                         )
                if retval != 0:
                    print "error running %r in %r. look in %r for details" %\
                          (cmd, os.getcwd(), logfile.name)

            dest_sp = os.path.join(dest_dir, "ubuntuone", "storageprotocol")
            shutil.copytree(os.path.join("ubuntuone", "storageprotocol"),
                            dest_sp)

        sp_revno = subprocess.check_output(["bzr", "revno", "."],
                                             env=BZR_ENV).strip()

        os.chdir(start_dir)

        # Get the revno for ubuntuone-client-data too
        os.chdir(os.path.join(self.source_dir, "ubuntuone-client-data"))
        u1cd_revno = subprocess.Popen(
            ["bzr", "revno", "."],
            stdout=subprocess.PIPE).communicate()[0]
        os.chdir(start_dir)
        # Copy the main icons file over
        shutil.copyfile(os.path.join(self.source_dir, "ubuntuone-client-data",
                                     "ubuntuone.icns"),
                        os.path.join("data", "ubuntuone.icns"))

        # Create revno file
        revno_str = ("ubuntu-sso-client: %s, "
                     "ubuntuone-client: %s, "
                     "ubuntuone-client-data: %s, "
                     "ubuntuone-control-panel: %s, "
                     "ubuntuone-storage-protocol: %s" %
                     (sso_revno, u1client_revno, u1cd_revno,
                      u1cp_revno, sp_revno))
        with open(os.path.join("data", "revnos.txt"), "w+") as revnos:
            revnos.write(revno_str)

        # Copy storage-protocol's pem files
        shutil.copyfile(os.path.join(self.source_dir,
            "ubuntuone-storage-protocol", "data",
            "UbuntuOne-Go_Daddy_CA.pem"),
            os.path.join("data", "UbuntuOne-Go_Daddy_CA.pem"))
        shutil.copyfile(os.path.join(self.source_dir,
            "ubuntuone-storage-protocol", "data",
            "UbuntuOne-Go_Daddy_Class_2_CA.pem"),
            os.path.join("data", "UbuntuOne-Go_Daddy_Class_2_CA.pem"))
        shutil.copyfile(os.path.join(self.source_dir,
                                     "ubuntuone-storage-protocol", "data",
                                     "UbuntuOne-ValiCert_Class_2_VA.pem"),
                        os.path.join("data",
                                     "UbuntuOne-ValiCert_Class_2_VA.pem"))

        # Copy syncdaemon config data
        shutil.copyfile(os.path.join(self.source_dir,
            "ubuntuone-client", "data",
            "syncdaemon.conf"),
            os.path.join("data", "syncdaemon.conf"))
        logging_path = os.path.join(self.source_dir,
            "ubuntuone-client", "data",
            "logging.conf.in")
        with open(logging_path, "rb") as logconf:
            data = logconf.read()
        data = data.replace("@LOG_LEVEL@", LOG_LEVEL)
        data = data.replace("@LOG_FILE_SIZE@", LOG_FILE_SIZE)
        with open(os.path.join("data", "logging.conf"), "wb") as logconf:
            logconf.write(data)

        print "building fsevents daemon"
        log_file_name = os.path.join(INSTALL_DIR,
                                     "fsevents-daemon-build.log")
        with open(log_file_name, 'w') as logfile:
            proj_path = os.path.join(self.source_dir,
                                     "ubuntuone-fsevents-daemon",
                                     "objc")
            cmd = ("cd %s && "
                   "xcodebuild -scheme FsEvents SYMROOT=%r clean build" %
                   (proj_path, INSTALL_DIR))
            retval = subprocess.call(cmd,
                                     shell=True,
                                     stderr=subprocess.STDOUT,
                                     stdout=logfile
                                     )
            if retval != 0:
                print "error running %r in %r. look in %r for details" %\
                    (cmd, os.getcwd(), logfile.name)

        # Install fsevents_daemon client code: required because it is
        # a subpackage of ubuntuone, which all needs to be in the same
        # place for py2app to copy it
        shutil.copyfile(os.path.join(self.source_dir,
                                     "ubuntuone-fsevents-daemon",
                                     "python",
                                     "ubuntuone",
                                     "fseventsd.py"),
                        os.path.join(dest_dir, "ubuntuone",
                                     "fseventsd.py"))

        print "building pyobjc cocoa menu"
        log_file_name = os.path.join(INSTALL_DIR,
                                     "cocoa-menu-build.log")
        with open(log_file_name, 'w') as logfile:
            proj_path = os.path.join(self.source_dir,
                                     "ubuntuone-cocoa-menu")
            file_abs_path = os.path.abspath(__file__)
            dist_path = os.path.join(os.path.dirname(file_abs_path),
                                     'dist')
            cmd = ("cd %s && "
                   "PYTHONPATH=$PYTHONPATH:%s/lib/site-packages/ "
                   "U1_SETUP_DIST_PATH=%s "
                   "python setup.py py2app -d %s" %
                   (proj_path, INSTALL_DIR, dist_path, dist_path))
            logfile.write("# %s\n" % cmd)
            retval = subprocess.call(cmd,
                                     shell=True,
                                     stderr=subprocess.STDOUT,
                                     stdout=logfile
                                     )
            if retval != 0:
                print "error running %r in %r. look in %r for details" %\
                    (cmd, os.getcwd(), logfile.name)

        # clear path_importer_cache to make sure that a following
        # py2app step can see the newly-copied
        # INSTALL_PATH/lib/site-packages/ubuntu_sso module without
        # clearing, py2app will fail on a clean run when both prepare
        # and py2app are specified on the same command:
        sys.path_importer_cache.clear()

        print "Prepare done"
# pylint: enable=R0904


if __name__ == '__main__':

    # this path is from homebrew:
    paths = glob.glob("/usr/local/Cellar/qt/*/plugins")
    if len(paths) != 1:
        print "Warning: expected just one path to Homebrew Qt plugins"
        print "but found ", paths

    qt_plugin_path = paths[0]

    qt_network_plugins = glob.glob(qt_plugin_path + "/bearer/*.dylib")
    qt_imageformat_plugins = glob.glob(qt_plugin_path +
                                       "/imageformats/*.dylib")
    qt_dylib_paths = qt_imageformat_plugins + qt_network_plugins

    libjpeg_paths = glob.glob("/usr/local/lib/libjpeg.*.dylib")
    if len(libjpeg_paths) != 1:
        print "Warning: expected to find just one version of libjpeg"
        print "found:", libjpeg_paths

    dylib_paths = qt_dylib_paths + [libjpeg_paths[0]]

    data_files = glob.glob("data/*")

    master_plist = Plist.fromFile('data/macapp_template.plist')

    master_options = {"includes": ['google.protobuf.descriptor',
                                   'sip',
                                   'twisted.web.resource',
                                   'twisted.web.client',
                                   'oauth',
                                   # workaround for indirect import in
                                   # utils/__init__.py:
                                   'ubuntu_sso.constants'
                                   ],
                      "packages": ["configglue"],
                      "excludes": ["fsm", "PyQt4.uic"],
                      "frameworks": dylib_paths,
                      "resources": data_files,
                      "strip": True,
                      }

    # make sure we can see the prepared libraries:
    sys.path.insert(0, os.path.join(INSTALL_DIR, "lib", "site-packages"))

    # start helper_app_paths with path to cocoa menu, which was built
    # in the prepare step:
    cocoa_menu_path = os.path.join('dist', 'Ubuntu One Menu.app')

    helper_app_paths = [cocoa_menu_path]

    def do_setup(is_background, id, exename,
                 verbose=False,
                 doprepare=False,
                 is_main_app=False):
        """Customize plist & options and run setup.

        - is_background controls showing a dock icon and menubar
        - id is the bundle id, in reverse dns style
        - exename is the name of the exe to be generated, usually
          should just be the basename of the main script without .py

        - verbose - sent to py2app, controls spewage
        - doprepare - set to True only the first call of setup()

        - is_main_app - if False, adds app path to list of helpers,
          which have their dependencies merged into the main app
          bundle in a post-processing step below.

        Returns the path of the generated .app bundle.
        """
        name = APP_NAMES[exename]
        mainscript = os.path.join("bin", exename + ".py")

        print "calling setup for:", name
        plist = copy.deepcopy(master_plist)

        if os.path.exists('data/revnos.txt'):
            revno_str = open('data/revnos.txt', 'r').read()
            ver_string = "%s (%s)" % (
                master_plist["CFBundleShortVersionString"],
                revno_str)
            plist.update(CFBundleShortVersionString=ver_string)

        plist.update(LSUIElement=is_background,
                     CFBundleIdentifier=id,
                     CFBundleExecutable=exename,
                     CFBundleName=name)

        if is_main_app:
            req_str = ("identifier %s and "
                       "certificate leaf[subject.CN] = \"%s\"" %
                       (FSEVENTS_DAEMON_NAME,
                        CODESIGN_CN))
            req_dict = {FSEVENTS_DAEMON_NAME: req_str}
            plist.update(SMPrivilegedExecutables=req_dict)

        options = copy.deepcopy(master_options)
        options.update({"plist": plist,
                        "bdist_base": os.path.join("build", exename),
                        "dist_dir":  os.path.join("dist", exename)})

        if doprepare:
            prepare_class = PrepareSources
        else:
            prepare_class = DoNotPrepareAnything

        setup(cmdclass=dict(prepare=prepare_class,
                            clean=Clean),
              app=[mainscript],
              options=dict(py2app=options),
              verbose=verbose)

        app_path = os.path.join("dist", exename, name + ".app")
        if not is_main_app:
            helper_app_paths.append(app_path)
        return app_path

    sso_login_qt_app_path = do_setup(True,
                                     "com.ubuntu.sso.login-qt",
                                     "ubuntu-sso-login-qt",
                                     doprepare=True)

    proxy_creds_app_path = do_setup(True,
                                    "com.ubuntu.sso.ssl-cert",
                                    "ubuntu-sso-proxy-creds-qt")

    sso_login_app_path = do_setup(True,
                                  "com.ubuntu.sso.login",
                                  "ubuntu-sso-login")

    proxy_tunnel_app_path = do_setup(True,
                                     "com.ubuntu.one.proxy-tunnel",
                                     "ubuntuone-proxy-tunnel")

    syncdaemon_app_path = do_setup(True,
                                   "com.ubuntu.one.syncdaemon",
                                   "ubuntuone-syncdaemon")

    u1sdtool_app_path = do_setup(True,
                                 "com.ubuntu.one.u1sdtool",
                                 "u1sdtool")

    updater_app_path = do_setup(True,
                                "com.ubuntu.one.updater",
                                 "ubuntuone-updater")

    control_panel_app_path = do_setup(False,
                                      "com.ubuntu.one.controlpanel",
                                      "ubuntuone-control-panel-qt",
                                      is_main_app=True)

    control_panel_contents_path = os.path.join(control_panel_app_path,
                                               "Contents")
    control_panel_resources_path = os.path.join(control_panel_contents_path,
                                                "Resources")

    # only do tweaks if you've done setup.py py2app
    if 'py2app' not in sys.argv:
        print "skipping post-setup tweaks."
        print "run setup.py py2app to build apps and finalize packaging"
        print "note that --source-dir only applies to 'prepare'."
        sys.exit()

    print "Starting post-setup tweaks"

    # move Qt plugins into Frameworks/plugins/subdir/
    control_panel_frameworks_path = os.path.join(control_panel_contents_path,
                                                 "Frameworks")

    plugin_names = [os.path.basename(f) for f in qt_dylib_paths]

    bearers_dir = os.path.join(control_panel_frameworks_path,
                               "plugins", "bearer")
    imagefmts_dir = os.path.join(control_panel_frameworks_path,
                                 "plugins", "imageformats")
    try:
        os.makedirs(imagefmts_dir)
    except Exception as e:
        print "ignoring exception in makedirs(%r): %r" % (imagefmts_dir,
                                                          e)

    try:
        os.makedirs(bearers_dir)
    except Exception as e:
        print "ignoring exception in makedirs(%r): %r" % (bearers_dir, e)

    try:
        for plugin_name in plugin_names:
            plugin_src = os.path.join(control_panel_frameworks_path,
                                      plugin_name)

            if "bearer" in plugin_name:
                plugin_dest = os.path.join(bearers_dir, plugin_name)
            else:
                plugin_dest = os.path.join(imagefmts_dir, plugin_name)

            os.rename(plugin_src, plugin_dest)

    except Exception as e:
        print "WARNING, got exception moving plugins around:", e

    def get_packages_zip_file(resources_path):
        """Get a ZipFile instance of the site_packages for an app."""
        zip_glob = os.path.join(resources_path,
                                'lib', 'python*',
                                'site-packages.zip')
        zip_paths = glob.glob(zip_glob)
        if len(zip_paths) != 1:
            raise Exception("zip_glob %r expanded to %r, wanted 1 path." %
                            (zip_glob, zip_paths))

        return ZipFile(zip_paths[0], 'a')

    def get_dynload_files(resources_path):
        """Get a set of files in the lib-dynload directory."""
        file_set = set()
        dynload_glob = os.path.join(resources_path,
                                    'lib', 'python*',
                                    'lib-dynload')
        dynload_paths = glob.glob(dynload_glob)
        if len(dynload_paths) != 1:
            raise Exception("dynload_glob %r expanded to %r, wanted 1 path." %
                            (dynload_glob, dynload_paths))

        dynload_path = dynload_paths[0]
        for root, dirs, files in os.walk(dynload_path):
            for f in files:
                path = os.path.join(root, f)
                relpath = os.path.relpath(path, dynload_path)
                file_set.add(relpath)
        return dynload_path, file_set

    main_zip = get_packages_zip_file(control_panel_resources_path)
    main_zip_files = set(main_zip.namelist())

    (main_dynload_path,
     main_dynload_files) = get_dynload_files(control_panel_resources_path)

    print "Copying helper apps into main app"

    for helper_path in helper_app_paths:

        print "Processing", helper_path

        helper_app_name = os.path.basename(helper_path)
        helper_src_contents_path = os.path.join(helper_path,
                                         "Contents")

        helper_dest = os.path.join(control_panel_resources_path,
                            helper_app_name)
        if os.path.exists(helper_dest):
            print "%r exists. removing it." % helper_dest
            shutil.rmtree(helper_dest)

        helper_dest_contents_path = os.path.join(helper_dest, "Contents")

        os.makedirs(helper_dest_contents_path)

        # copy actual Contents/MacOS and plist, symlink everything else

        shutil.copytree(os.path.join(helper_src_contents_path, "MacOS"),
                        os.path.join(helper_dest_contents_path, "MacOS"))

        shutil.copy(os.path.join(helper_src_contents_path, "Info.plist"),
                    os.path.join(helper_dest_contents_path, "Info.plist"))

        # link to the rest of the items in the main app bundle, except:
        # - the .app bundles, to avoid circular links
        # __boot__.py needs to be the one from the subapp

        resource_subdirs_to_link = ['include', 'lib']

        helper_src_resources_path = os.path.join(helper_src_contents_path,
                                                 "Resources")

        helper_dest_resources_path = os.path.join(helper_dest_contents_path,
                                                  "Resources")
        os.mkdir(helper_dest_resources_path)

        helper_src_resources = os.listdir(helper_src_resources_path)

        # copy everything in resources that isn't include or lib
        resources_to_copy = [p for p in helper_src_resources if
                             p not in resource_subdirs_to_link]

        for rsrc_name in resources_to_copy:
            src_rsrc_path = os.path.join(helper_src_resources_path, rsrc_name)
            dest_rsrc_path = os.path.join(helper_dest_resources_path,
                                          rsrc_name)
            if os.path.isdir(src_rsrc_path):
                if rsrc_name == 'English.lproj':
                    shutil.copytree(src_rsrc_path, dest_rsrc_path)
                else:
                    print "WARNING: ignoring unexpected source directory",
            else:
                shutil.copy(src_rsrc_path, dest_rsrc_path)

        # link to resources/include and resources/lib from the main bundle:
        resources_to_link = [os.path.join("Resources", p)
                             for p in resource_subdirs_to_link]

        # also link Frameworks directory
        for src_item in ["Frameworks", "PkgInfo"] + resources_to_link:
            src_path = os.path.join(os.path.pardir,
                                    os.path.pardir,
                                    os.path.pardir,
                                    src_item)
            # add another pardir to make Resources/* items work:
            if src_item.startswith("Resources"):
                src_path = os.path.join(os.path.pardir, src_path)

            link_name = os.path.join(helper_dest_contents_path, src_item)
            os.symlink(src_path, link_name)

        print "Merging site-packages.zip files"
        helper_zip = get_packages_zip_file(helper_src_resources_path)
        helper_zip_files = set(helper_zip.namelist())
        add_from_helper = helper_zip_files - main_zip_files

        for name in add_from_helper:
            main_zip.writestr(name, helper_zip.read(name))
        helper_zip.close()

        print "Merging lib-dynload contents"
        (helper_dynload_path,
         helper_dynload_files) = get_dynload_files(helper_src_resources_path)
        add_from_helper = helper_dynload_files - main_dynload_files

        for filename in add_from_helper:
            src = os.path.join(helper_dynload_path,
                               filename)
            dest = os.path.join(main_dynload_path,
                                filename)
            dest_dir = os.path.dirname(dest)
            if not os.path.exists(dest_dir):
                os.makedirs(dest_dir)
            shutil.copy(src, dest)

    main_zip.close()

    print "copying translation files"
    shutil.copytree(os.path.join(INSTALL_DIR, 'share', 'locale'),
                    os.path.join(control_panel_resources_path,
                                 'translations'))

    print "building signing package"
    sign_dir = os.path.join("dist", "build_to_sign")
    os.makedirs(sign_dir)

    shutil.move(control_panel_app_path, sign_dir)
    daemon_build_path = os.path.join(INSTALL_DIR, "Debug",
                                     FSEVENTS_DAEMON_NAME)
    shutil.copy(daemon_build_path, sign_dir)
    shutil.copy(os.path.join("codesign-darwin.sh"),
                sign_dir)
    shutil.copy(os.path.join("codesign-darwin-verify.sh"),
                sign_dir)

    print "Done. See %r for the signing package." % sign_dir
